# coding=utf-8
from __future__ import print_function

import pandas as pd

from suanpan import path


def load(file, *args, **kwargs):
    kwargs.setdefault("index_col", 0)
    return pd.read_csv(file, *args, **kwargs)


def dump(dataframe, file, *args, **kwargs):
    path.safeMkdirsForFile(file)
    dataframe.to_csv(file, **kwargs)
    return file



# coding=utf-8
from __future__ import print_function

import abc
import argparse

from suanpan import objects
from suanpan.dw import dw
from suanpan.log import logger
from suanpan.mq import mq
from suanpan.mstorage import mstorage
from suanpan.storage import storage


class HasBaseServices(abc.ABC):
    ENABLED_BASE_SERVICES = {"mq", "mstorage", "dw", "storage"}

    @classmethod
    def setBaseServices(cls, args):
        mapping = {
            "mq": cls.setMQ,
            "mstorage": cls.setMStorage,
            "dw": cls.setDataWarehouse,
            "storage": cls.setStorage,
        }
        setters = [mapping.get(s) for s in cls.ENABLED_BASE_SERVICES]
        if not all(setters):
            raise Exception(
                "Unknown base serivces: {}".format(
                    set(cls.ENABLED_BASE_SERVICES) - set(mapping.keys())
                )
            )
        for setter in setters:
            setter(args)

    @classmethod
    def setMQ(cls, args):
        return mq.setBackend(**cls.defaultArgumentsFormat(args, mq.ARGUMENTS))

    @classmethod
    def setMStorage(cls, args):
        return mstorage.setBackend(
            **cls.defaultArgumentsFormat(args, mstorage.ARGUMENTS)
        )

    @classmethod
    def setDataWarehouse(cls, args):
        return dw.setBackend(**cls.defaultArgumentsFormat(args, dw.ARGUMENTS))

    @classmethod
    def setStorage(cls, args):
        return storage.setBackend(**cls.defaultArgumentsFormat(args, storage.ARGUMENTS))

    @abc.abstractmethod
    def getArguments(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def transformArguments(self, *args, **kwargs):
        pass

    def parseArgs(self):
        return self.parseArguments(self.getArguments())

    @classmethod
    def parseArguments(cls, arguments, *args, **kwargs):
        parser = argparse.ArgumentParser(*args, **kwargs)
        for arg in arguments:
            arg.addParserArguments(parser)
        return parser.parse_known_args()[0]

    @classmethod
    def loadArguments(cls, args, arguments):
        return {arg.key: arg.load(args) for arg in arguments}

    @classmethod
    def formatArguments(cls, context, arguments):
        return {arg.key: arg.format(context) for arg in arguments}

    @classmethod
    def cleanArguments(cls, context, arguments):
        return {arg.key: arg.clean(context) for arg in arguments}

    @classmethod
    def defaultArgumentsFormat(cls, args, arguments):
        arguments = (arg.key.replace("-", "_") for arg in arguments)
        return {
            cls.defaultArgumentKeyFormat(arg): getattr(args, arg, None) for arg in arguments
        }

    @classmethod
    def defaultArgumentKeyFormat(cls, key):
        return cls.toCamelCase(cls.removePrefix(key))

    @classmethod
    def removePrefix(cls, string, delimiter="_", num=1):
        pieces = string.split(delimiter)
        pieces = pieces[num:] if len(pieces) > num else pieces
        return delimiter.join(pieces)

    @classmethod
    def toCamelCase(cls, string, delimiter="_"):
        camelCaseUpper = lambda i, s: s[0].upper() + s[1:] if i and s else s
        return "".join(
            [camelCaseUpper(i, s) for i, s in enumerate(string.split(delimiter))]
        )


class HasLogger(objects.HasName):
    def __init__(self):
        super(HasLogger, self).__init__()
        logger.setLogger(self.name)

    @property
    def logger(self):
        return logger



# coding=utf-8
from __future__ import print_function

import argparse
import contextlib
import copy
import itertools
import os
import re
import sys
import traceback
from collections import defaultdict

from suanpan import g, objects
from suanpan.arguments import Bool
from suanpan.log import logger


class Arguments(objects.Context):
    pass


class Result(objects.Context):
    def __init__(self, value):
        super(Result, self).__init__(value=value)


class Component(object):

    ENVIRONMENT_ARGUMENTS_KEYWORD = "SP_PARAM"

    devArguments = [Bool(key="debug", default=False)]
    defaultArguments = []

    def __init__(self, funcOrComponent):
        if isinstance(funcOrComponent, Component):
            self.runFunc = funcOrComponent.runFunc
            self.arguments = funcOrComponent.arguments
        else:
            self.runFunc = funcOrComponent
            self.arguments = defaultdict(list)

    def __call__(self, *arg, **kwargs):
        try:
            self.run(*arg, **kwargs)
        except Exception as e:
            logger.error(traceback.format_exc())
            raise e

    @property
    def name(self):
        return self.runFunc.__name__

    def run(self, *arg, **kwargs):
        logger.info("Starting...")
        context = self.init()
        results = self.runFunc(context)
        self.save(context, results)
        logger.info("Done.")

    def init(self):
        restArgs = self.getArgList()
        devArgs, restArgs = self.parseDevArguments(restArgs)
        g.update(vars(devArgs))
        defaultArgs, restArgs = self.parseDefaultArguments(restArgs)
        self.initBase(defaultArgs)
        context = self.getContext(defaultArgs)
        args, restArgs = self.parseArguments(self.getArguments(), restArgs)
        args = self.transformArguments(context, args)
        setattr(context, "args", args)
        return context

    def initBase(self, args):
        pass

    def save(self, context, results):
        outputs = self.saveOutputs(context, results)
        self.closeContext()
        return outputs

    @contextlib.contextmanager
    def context(self, args=None):
        yield objects.Context()

    def getContext(self, *args, **kwargs):
        self.contextManager = self.context(*args, **kwargs)
        return next(self.contextManager.gen)  # pylint: disable-msg=e1101

    def closeContext(self):
        try:
            next(self.contextManager.gen)  # pylint: disable-msg=e1101
        except StopIteration:
            pass

    def parseDevArguments(self, argList=None):
        return self.parseArguments(
            self.devArguments, argList=argList, description="Dev arguments"
        )

    def parseDefaultArguments(self, argList=None):
        return self.parseArguments(
            self.defaultArguments, argList=argList, description="Default arguments"
        )

    def getArgList(self):
        return self.getArgListFromEnv() or self.getArgListFromCli()

    def getArgListFromCli(self):
        return sys.argv[1:]

    def getArgListFromEnv(self):
        argString = os.environ.get(self.ENVIRONMENT_ARGUMENTS_KEYWORD, "")
        regex = r"(--[\w-]+)\s+(?:(?P<quote>[\"\'])(.*?)(?P=quote)|([^\'\"\s]+))"
        groups = re.findall(regex, argString)
        argv = list(
            itertools.chain(*[(group[0], group[-2] or group[-1]) for group in groups])
        )
        return argv

    def parseArguments(self, arguments, argList=None, **kwargs):
        parser = argparse.ArgumentParser(**kwargs)
        for arg in arguments:
            arg.addParserArguments(parser)
        return parser.parse_known_args(argList)

    def transformArguments(self, context, args):
        self.loadArguments(args, self.getArguments())
        self.formatArguments(context, self.getArguments(exclude="outputs"))
        self.cleanArguments(context, self.getArguments(include="outputs"))
        arguments = {
            k: {arg.key: arg.value for arg in v if arg.isSet}
            for k, v in self.arguments.items()
        }
        arguments.update(
            {arg.key: arg.value for arg in itertools.chain(*self.arguments.values())}
        )
        return Arguments(**arguments)

    def loadArguments(self, args, arguments):
        return {arg.key: arg.load(args) for arg in arguments}

    def _safe_list(self, params=None):
        return [params] if isinstance(params, str) else list(params)

    def getArguments(self, include=None, exclude=None):
        includes = set(
            self.arguments.keys() if not include else self._safe_list(include)
        )
        excludes = set([] if not exclude else self._safe_list(exclude))
        includes = includes - excludes
        argumentsLists = [self.arguments[c] for c in includes]
        return itertools.chain(*argumentsLists)

    def formatArguments(self, context, arguments):
        return {arg.key: arg.format(context) for arg in arguments}

    def cleanArguments(self, context, arguments):
        return {arg.key: arg.clean(context) for arg in arguments}

    def saveMutipleOutputs(self, context, outputs, results):
        if isinstance(results, tuple) or isinstance(results, list):
            return {
                argument.key: argument.save(context, Result(result))
                for argument, result in zip(outputs, results)
            }
        elif isinstance(results, dict):
            return {
                argument.key: argument.save(context, Result(results.get(argument.key)))
                for argument in outputs
                if results.get(argument.key) is not None
            }
        else:
            raise Exception("Incorrect results: {}".format(results))

    def saveOneOutput(self, context, output, results):
        result = (
            Result(results[output.key])
            if isinstance(results, dict) and output.key in results
            else Result(results)
        )
        return {output.key: output.save(context, result)}

    def saveOutputs(self, context, results):
        logger.info("Saving...")
        outputs = list(self.getArguments("outputs"))
        if len(outputs) > 1:
            return self.saveMutipleOutputs(context, outputs, results)
        elif len(outputs) == 1:
            return self.saveOneOutput(context, outputs[0], results)

    def addArgument(self, arg, argtype="args", reverse=True):
        if reverse:
            self.arguments[argtype].insert(0, arg)
        else:
            self.arguments[argtype].append(arg)

    @classmethod
    def arg(cls, argument, argtype="args", reverse=True):
        def _dec(funcOrComponent):
            funcOrComponent = (
                funcOrComponent
                if isinstance(funcOrComponent, cls)
                else cls(funcOrComponent)
            )
            funcOrComponent.addArgument(argument, argtype=argtype, reverse=reverse)
            return funcOrComponent

        return _dec

    @classmethod
    def input(cls, *args, **kwargs):
        kwargs.update(argtype="inputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def output(cls, *args, **kwargs):
        kwargs.update(argtype="outputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def param(cls, *args, **kwargs):
        kwargs.update(argtype="params")
        return cls.arg(*args, **kwargs)

    @classmethod
    def column(cls, *args, **kwargs):
        kwargs.update(argtype="columns")
        return cls.arg(*args, **kwargs)



# coding=utf-8
from __future__ import print_function

from suanpan.arguments import *  # pylint: disable-msg=w0614
from suanpan.dw.arguments import *  # pylint: disable-msg=w0614
from suanpan.storage.arguments import *  # pylint: disable-msg=w0614



# coding=utf-8
from __future__ import print_function

import contextlib
import functools
import multiprocessing
from multiprocessing import Pool as ProcessPool
from multiprocessing import current_process as currentProcess
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing.dummy import current_process as currentThread

import tqdm

from suanpan import utils

WORKERS = multiprocessing.cpu_count()
DEFAULT_PBAR_FORMAT = "{desc}: {n_fmt}/{total_fmt} |{bar}"
DEFAULT_PBAR_CONFIG = {"bar_format": DEFAULT_PBAR_FORMAT}


@contextlib.contextmanager
def multiThread(workers=None):
    workers = workers or WORKERS
    pool = ThreadPool(processes=workers)
    yield pool
    pool.close()
    pool.join()


@contextlib.contextmanager
def multiProcess(workers=None):
    workers = workers or WORKERS
    pool = ProcessPool(processes=workers)
    yield pool
    pool.close()
    pool.join()


def parsePbarConfig(pbar):
    if pbar is True:
        return utils.merge({}, DEFAULT_PBAR_CONFIG, {"desc": "Processing"})
    elif isinstance(pbar, str):
        return utils.merge({}, DEFAULT_PBAR_CONFIG, {"desc": pbar})
    elif pbar in (False, None):
        return utils.merge({}, DEFAULT_PBAR_CONFIG, {"disable": True})
    elif isinstance(pbar, dict):
        return utils.merge({}, DEFAULT_PBAR_CONFIG, pbar)
    else:
        raise Exception("Invalid pbar config: bool | str | dict. but {}".format(pbar))


def pbarRunner(pbar, quantity=1):
    def _dec(runner):
        @functools.wraps(runner)
        def _runner(*args, **kwargs):
            result = runner(*args, **kwargs)
            pbar.update(quantity)
            return result

        return _runner

    return _dec


def _map(
    func, iterable, chunksize=None, workers=None, pbar=None, thread=False, mapFunc="map"
):
    items = list(iterable)
    pbarConfig = parsePbarConfig(pbar)
    pbarConfig.update(total=len(items))
    poolClass = multiThread if thread else multiProcess
    with poolClass(workers) as pool:
        with tqdm.tqdm(**pbarConfig) as pbar:
            run = pbarRunner(pbar)(func)
            mapFunc = getattr(pool, mapFunc)
            results = mapFunc(run, items, chunksize)
    return results


map = functools.partial(_map, mapFunc="map")
starmap = functools.partial(_map, mapFunc="starmap")
imap = functools.partial(_map, mapFunc="imap")



# coding=utf-8
from __future__ import print_function

from suanpan.arguments import Bool, Int, String
from suanpan.mq.redis import RedisMQ
from suanpan.proxy import Proxy


class MQProxy(Proxy):
    MAPPING = {"redis": RedisMQ}
    ARGUMENTS = [
        String("mq-type", default="redis"),
        String("mq-redis-host", default="localhost"),
        Int("mq-redis-port", default=6379),
        Bool("mq-redis-realtime", default=False),
    ]


mq = MQProxy()



# coding=utf-8
from __future__ import print_function

from tabulate import tabulate

# from termgraph import termgraph


# def chart(data, labels=None, colors=None, **kwargs):
#     args = {
#         "filename": "-",
#         "title": None,
#         "width": 50,
#         "format": "{:<5.2f}",
#         "suffix": "",
#         "no_labels": not bool(labels),
#         "color": None,
#         "vertical": False,
#         "stacked": False,
#         "different_scale": False,
#         "calendar": False,
#         "start_dt": None,
#         "custom_tick": "",
#         "delim": "",
#         "verbose": False,
#         "version": False,
#     }
#     args.update(kwargs)
#     colors = colors or []
#     labels = labels or [None] * len(data)
#     if args["calendar"]:
#         termgraph.calendar_heatmap(data, labels, args)
#     else:
#         termgraph.chart(colors, data, args, labels)


def table(data, headers=(), theme="plain"):
    print(tabulate(data, headers=headers, tablefmt=theme))



# coding=utf-8
from __future__ import print_function

import itertools

from suanpan import interfaces
from suanpan.components import Component
from suanpan.dw import dw
from suanpan.log import logger
from suanpan.storage import storage


class DockerComponent(Component, interfaces.HasLogger, interfaces.HasBaseServices):
    ENABLED_BASE_SERVICES = {"dw", "storage"}
    defaultArguments = list(itertools.chain(dw.ARGUMENTS, storage.ARGUMENTS))

    def initBase(self, args):
        logger.setLogger(self.name)
        self.setBaseServices(args)
